from .health_technology_template import HealthTechnologies, HealthTechnology


health_technologies = HealthTechnologies(
    hypertension_drugs=HealthTechnology(
        name='hypertension_drugs',
        kind='health_technology',
        gbd_id=None,
    )
)
